from datetime import datetime
from typing import Iterable, List, Optional

from graia.broadcast.typing import T_Dispatcher

Timer = Iterable[datetime]

from asyncio import AbstractEventLoop
import asyncio

from graia.broadcast import Broadcast
from graia.broadcast.entities.decorator import Decorator

from .task import SchedulerTask
from typing import Callable, TypeVar

T_Callable = TypeVar("T_Callable", bound=Callable)


class GraiaScheduler:
    """任务计划器"""

    loop: AbstractEventLoop
    schedule_tasks: List[SchedulerTask]
    broadcast: Broadcast

    def __init__(self, loop: AbstractEventLoop, broadcast: Broadcast) -> None:
        """初始化

        Args:
            loop (AbstractEventLoop): 事件循环
            broadcast (Broadcast): 事件总线
        """
        self.schedule_tasks = []
        self.loop = loop
        self.broadcast = broadcast

    def schedule(
        self,
        timer: Timer,
        cancelable: bool = False,
        dispatchers: Optional[List[T_Dispatcher]] = None,
        decorators: Optional[List[Decorator]] = None,
    ) -> Callable[[T_Callable], T_Callable]:
        """计划一个新任务.

        Args:
            timer (Timer): 该定时任务的计时器.
            cancelable (bool, optional): 能否取消该任务. 默认为 False.
            dispatchers (List[T_Dispatcher], optional): 该任务要使用的 Dispatchers. 默认为空列表.
            decorators (Optional[List[Decorator]], optional): 该任务要使用的 Decorators. 默认为空列表.

        Returns:
            Callable[[T_Callable], T_Callable]: 任务 函数/方法 包装器.
        """

        def wrapper(func):
            task = SchedulerTask(
                func,
                timer,
                self.broadcast,
                self.loop,
                cancelable,
                dispatchers,
                decorators,
            )
            self.schedule_tasks.append(task)
            return func

        return wrapper

    async def run(self) -> None:
        """开始所有计划任务, 在所有任务结束后返回"""
        await asyncio.gather(*(task.setup_task() for task in self.schedule_tasks))

    async def join(self, stop: bool = False) -> None:
        """等待所有计划任务结束"""
        await asyncio.gather(*(task.join(stop=stop) for task in self.schedule_tasks))

    def stop(self) -> None:
        """停止所有计划任务"""
        for task in self.schedule_tasks:
            task.stop()



from dataclasses import dataclass, field
from typing import List

from graia.broadcast.entities.decorator import Decorator
from graia.broadcast.typing import T_Dispatcher
from graia.saya.schema import BaseSchema

from .. import Timer


@dataclass
class SchedulerSchema(BaseSchema):
    timer: Timer
    cancelable: bool = field(default=False)
    decorators: List[Decorator] = field(default_factory=list)
    dispatchers: List[T_Dispatcher] = field(default_factory=list)



from typing import Literal, Set
from launart import Launchable, Launart
from . import GraiaScheduler
import asyncio


class SchedulerService(Launchable):
    """GraiaScheduler 的 Launart 服务

    Args:
        scheduler (GraiaScheduler): 任务计划器
    """

    id = "scheduler.service"

    def __init__(self, scheduler: GraiaScheduler) -> None:
        super().__init__()
        self.scheduler = scheduler

    @property
    def required(self):
        return set()

    @property
    def stages(self) -> Set[Literal["preparing", "blocking", "cleanup"]]:
        return {"preparing", "cleanup"}

    async def launch(self, manager: Launart):
        async with self.stage("preparing"):
            pass  # Wait for preparation to complete, then we run tasks

        tsk = asyncio.create_task(self.scheduler.run())

        async with self.stage("cleanup"):
            # Stop all schedule tasks
            self.scheduler.stop()
            await self.scheduler.join()
            await tsk



from datetime import datetime, time
from typing import Union


class EnteredRecord:
    def __init__(self) -> None:
        self.entered = False

    def __enter__(self) -> None:
        self.entered = True

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.entered = False


def print_track_async(func):
    async def wrapper(*args, **kwargs):
        try:
            return await func(*args, **kwargs)
        except:
            import traceback

            traceback.print_exc()

    return wrapper


TimeObject = Union[datetime, time, str, float]


def to_datetime(
    base: TimeObject,
) -> datetime:
    """将适宜的对象转化为 datetime 类型.

    Args:
        base (TimeObject): 要转化的对象, 字符串应为 ISO 时间 / 日期格式, 浮点数为时间戳.

    Raises:
        ValueError: 字符串格式错误.

    Returns:
        datetime: 转化成的 datetime.
    """
    start = datetime.now()
    if isinstance(base, datetime):
        return base
    elif isinstance(base, time):
        return datetime.fromisoformat(f"{start.date()} {base.isoformat()}")
    elif isinstance(base, float):
        return datetime.fromtimestamp(base)
    else:
        if "-" in base:  # Base is from datetime.isoformat()
            return datetime.fromisoformat(base)
        elif ":" in base:  # Base is from time.isoformat()
            return datetime.fromisoformat(f"{start.date()} {base}")
        else:
            raise ValueError("Expected an ISO style time string!")



from typing import Any, Union

from graia.saya.behaviour import Behaviour
from graia.saya.cube import Cube

from graia.scheduler import GraiaScheduler
from graia.scheduler.saya.schema import SchedulerSchema


class GraiaSchedulerBehaviour(Behaviour):
    scheduler: GraiaScheduler

    def __init__(self, scheduler: GraiaScheduler) -> None:
        self.scheduler = scheduler

    def allocate(self, cube: Cube[SchedulerSchema]):
        if isinstance(cube.metaclass, SchedulerSchema):
            self.scheduler.schedule(
                cube.metaclass.timer,
                cube.metaclass.cancelable,
                cube.metaclass.dispatchers,
                cube.metaclass.decorators,
            )(cube.content)
        else:
            return

        return True

    def release(self, cube: Cube) -> Any:
        if isinstance(cube.metaclass, SchedulerSchema):
            target_tasks = list(
                filter(
                    lambda x: x.target is cube.content, self.scheduler.schedule_tasks
                )
            )
            if target_tasks:
                target = target_tasks[0]
                target.stop_gen_interval()
                target.stop()
                self.scheduler.schedule_tasks.remove(target)
        else:
            return

        return True

    uninstall = release


"""该模块提供一些便捷的 Timer"""

from datetime import datetime, timedelta
from typing import Generator, Optional

from croniter import croniter

from graia.scheduler.utilles import TimeObject, to_datetime

Ret = Generator[datetime, None, None]


def every(*, base: Optional[TimeObject] = None, **kwargs) -> Ret:
    """一个简便的 datetime 生成器.

    Args:
        base (Optional[TimeObject], optional): 若为 None (默认), 则会相对于当前时间推算. 否则基于 base 推算.

    Yields:
        (datetime): 生成的 datetime.
    """
    if base is None:
        while True:
            yield datetime.now() + timedelta(**kwargs)
    else:
        current = to_datetime(base)
        while True:
            current += timedelta(**kwargs)
            yield current


def every_second(*, base: Optional[TimeObject] = None) -> Ret:
    """每秒钟执行一次

    Args:
        base (Optional[TimeObject], optional): 若为 None (默认), 则会相对于当前时间推算. 否则基于 base 推算.

    Yields:
        (datetime): 生成的 datetime.
    """
    yield from every(seconds=1, base=base)


def every_minute(*, base: Optional[TimeObject] = None) -> Ret:
    """每分钟执行一次.

    Args:
        base (Optional[TimeObject], optional): 若为 None (默认), 则会相对于当前时间推算. 否则基于 base 推算.

    Yields:
        (datetime): 生成的 datetime.
    """
    yield from every(minutes=1, base=base)


def every_hour(*, base: Optional[TimeObject] = None):
    """每小时执行一次.

    Args:
        base (Optional[TimeObject], optional): 若为 None (默认), 则会相对于当前时间推算. 否则基于 base 推算.

    Yields:
        (datetime): 生成的 datetime.
    """
    yield from every(hours=1, base=base)


every_hours = every_hour  # Backward compatibility


def every_custom_seconds(seconds: int, *, base: Optional[TimeObject] = None) -> Ret:
    """每 seconds 秒执行一次

    Args:
        seconds (int): 距离下一次执行的时间间隔, 单位为秒
        base (Optional[TimeObject], optional): 若为 None (默认), 则会相对于当前时间推算. 否则基于 base 推算.

    Yields:
        (datetime): 生成的 datetime.
    """
    yield from every(seconds=seconds, base=base)


def every_custom_minutes(minutes: int, *, base: Optional[TimeObject] = None) -> Ret:
    """每 minutes 分执行一次

    Args:
        minutes (int): 距离下一次执行的时间间隔, 单位为分
        base (Optional[TimeObject], optional): 若为 None (默认), 则会相对于当前时间推算. 否则基于 base 推算.

    Yields:
        (datetime): 生成的 datetime.
    """
    yield from every(minutes=minutes, base=base)


def every_custom_hours(hours: int, *, base: Optional[TimeObject] = None) -> Ret:
    """每 hours 小时执行一次

    Args:
        hours (int): 距离下一次执行的时间间隔, 单位为小时
        base (Optional[TimeObject], optional): 若为 None (默认), 则会相对于当前时间推算. 否则基于 base 推算.

    Yields:
        (datetime): 生成的 datetime.
    """
    yield from every(hours=hours, base=base)


def crontabify(pattern: str, base: Optional[TimeObject] = None) -> Ret:
    """使用类似 crontab 的方式生成计时器

    Args:
        pattern (str): 时间模式
        base (Optional[TimeObject], optional): 开始时间. 默认为 datetime.now().

    Yields:
        (datetime): 生成的 datetime.
    """
    base = to_datetime(base) if base else datetime.now()
    crontab_iter = croniter(pattern, base)
    while True:
        yield crontab_iter.get_next(datetime)



import asyncio
import traceback
from datetime import datetime
from typing import Any, Callable, Generator, List, Optional, Tuple, Awaitable

from graia.broadcast import Broadcast
from graia.broadcast.entities.decorator import Decorator
from graia.broadcast.entities.exectarget import ExecTarget
from graia.broadcast.exceptions import ExecutionStop, PropagationCancelled
from graia.broadcast.typing import T_Dispatcher
from graia.broadcast.builtin.event import ExceptionThrown
from graia.scheduler.exception import AlreadyStarted
from graia.scheduler.utilles import EnteredRecord, print_track_async

from . import Timer


class SchedulerTask:
    target: Callable[..., Any]
    timer: Timer
    task: Optional[asyncio.Task]

    broadcast: Broadcast
    dispatchers: List[T_Dispatcher]
    decorators: List[Decorator]

    cancelable: bool
    stopped: bool

    sleep_record: EnteredRecord
    run_record: EnteredRecord

    loop: asyncio.AbstractEventLoop

    @property
    def is_sleeping(self) -> bool:
        return self.sleep_record.entered

    @property
    def is_executing(self) -> bool:
        return self.run_record.entered and not self.sleep_record.entered

    def __init__(
        self,
        target: Callable[..., Any],
        timer: Timer,
        broadcast: Broadcast,
        loop: Optional[asyncio.AbstractEventLoop] = None,
        cancelable: bool = False,
        dispatchers: Optional[List[T_Dispatcher]] = None,
        decorators: Optional[List[Decorator]] = None,
    ) -> None:
        self.target = target
        self.timer = timer
        self.broadcast = broadcast
        self.loop = loop or asyncio.get_running_loop()
        self.cancelable = cancelable
        self.task = None
        self.stopped = False
        self.dispatchers = dispatchers or []
        self.decorators = decorators or []
        self.sleep_record = EnteredRecord()
        self.run_record = EnteredRecord()

    def setup_task(self) -> asyncio.Task:
        """将本 SchedulerTask 作为 asyncio.Task 排入事件循环."""
        if self.task:
            raise AlreadyStarted("the scheduler task has been started!")
        self.task = self.loop.create_task(self.run())
        return self.task

    def sleep_interval_generator(self) -> Generator[float, None, None]:
        for next_execute_time in self.timer:
            if self.stopped:
                return
            now = datetime.now()
            if next_execute_time >= now:
                yield (next_execute_time - now).total_seconds()

    def coroutine_generator(self) -> Generator[Tuple[Awaitable[Any], bool], None, None]:
        for sleep_interval in self.sleep_interval_generator():
            yield (asyncio.sleep(sleep_interval), True)
            yield (
                self.broadcast.Executor(
                    target=ExecTarget(
                        callable=self.target,
                        inline_dispatchers=self.dispatchers,
                        decorators=self.decorators,
                    ),
                ),
                False,
            )

    @print_track_async
    async def run(self) -> None:
        if self.run_record.entered:
            raise AlreadyStarted("the scheduler task has been started!")
        with self.run_record:
            for coro, waiting in self.coroutine_generator():
                if waiting:  # 是否为 asyncio.sleep 的 coro
                    with self.sleep_record:
                        try:
                            await coro
                        except asyncio.CancelledError:
                            return
                try:
                    await (coro if self.cancelable else asyncio.shield(coro))
                except asyncio.CancelledError:
                    if self.cancelable:
                        return
                    raise
                except (ExecutionStop, PropagationCancelled):
                    pass
                except Exception as e:
                    traceback.print_exc()
                    await self.broadcast.postEvent(ExceptionThrown(e, None))

    def stop_gen_interval(self) -> None:
        if not self.stopped:
            self.stopped = True

    async def join(self, stop: bool = False) -> None:
        """阻塞直至当前 SchedulerTask 执行完毕.

        Args:
            stop (bool, optional): 是否停止当前 SchedulerTask 下一次运行. 默认为 False.
        """
        if stop and not self.stopped:
            self.stop_gen_interval()

        if self.task:
            await self.task
        self.task = None

    def stop(self):
        """停止当前 SchedulerTask."""
        if self.task and not self.task.cancelled():
            self.task.cancel()
